#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <math.h>
#include <conio.h>

#define  ERR_NONE          0
#define  ERR_DPMI_FAIL     1
#define  ERR_NO_VESA       2
#define  ERR_MODE          3
#define  ERR_MEM           4

#ifndef ABS
#define  ABS(a)            ((a) < 0 ? -(a) : (a))
#endif

// choices for memory model
#define  VESA_BANKED     0x201  // force banked (default)
#define  VESA_LFB        0x202  // force LFB's  (uses banked if no LFB's)

#define  PUT_ALL         0x001
#define  PUT_TRANS       0x002

#define  RMFP_LINEAR(a)    ((((a) >> 16) << 4) + ((a) & 0xffff))

#define  TO_15BIT(r, g, b) (((unsigned int)b >> 3) | (((unsigned int)g >> 3) << 5) | (((unsigned int)r >> 3) << 10))
#define  TO_16BIT(r, g, b) (((unsigned int)b >> 3) | (((unsigned int)g >> 2) << 5) | (((unsigned int)r >> 3) << 11))
#define  TO_32BIT(r, g, b) (b | ((unsigned int)g << 8) | ((unsigned int)r << 16))

typedef struct
{
  int  Height,Width;
  int  BitsPerPixel;
  char Data[];
} Sprite;

typedef struct
{
  int Mode;
  int XRes;
  int YRes;
  int CDepth;
} VESAMode;

typedef struct
{
  char       VESASignature[4];
  short int  VESAVersion;
  char       *OEMStringPtr;
  int        Capabilities;
  int        VideoModePtr;
  short int  TotalMemory;
  short int  OemSoftwareRev;
  char       *OemVendorNamePtr;
  char       *OemProductNamePtr;
  char       *OemProductRevPtr;
  char       Reserved[222];
  char       OemData[256];
} VESAInfoBlock;

typedef struct
{
  short int  ModeAttributes;
  char       WinAAttributes;
  char       WinBAttributes;
  short int  WinGranularity;
  short int  WinSize;
  short int  WinASegment;
  short int  WinBSegment;
  int        WinFuncPtr;
  short int  BytesPerLine;
  short int  XResolution;
  short int  YResolution;
  char       XCharSize;
  char       YCharSize;
  char       NumberOfPlanes;
  char       BitsPerPixel;
  char       NumberOfBanks;
  char       MemoryModel;
  char       BankSize;
  char       NumberOfPages;
  char       Reserved1;
  char       RedMaskSize;
  char       RedFieldPosition;
  char       GreenMaskSize;
  char       GreenFieldPosition;
  char       BlueMaskSize;
  char       BlueFieldPosition;
  char       RsvdMaskSize;
  char       RsvdFieldPosition;
  char       DirectColorModeInfo;
  int        PhysBasePtr;
  int        OffScreenMemOffset;
  short int  OffScreenMemSize;
  char       Reserved[206];
} ModeInfoBlock;


typedef struct
{
    int maxx, maxy;
    int linesize;
    int currentmode;
    int colordepth;
    int granularity;
    int granmask;
    int granshift;
    int pagecache;
    int ingraphics;
    int isvesa;
    int memmode;
    int bytesperpixel;

    int redsizeshift;
    int greensizeshift;
    int bluesizeshift;
    int redlocshift;
    int greenlocshift;
    int bluelocshift;

    int version;
    int errorflag;
    int nummodes;
    int totalmemory;

    VESAMode *VESAModes;
    VESAInfoBlock *InfoBlock;
    ModeInfoBlock *ModeBlock;

    char *VGAScreen;
} VESACurrentModeBlock;



void RMint86(char intnum, union REGS *inregs, union REGS *outregs);
void RMint86x(char intnum, union REGS *inregs, union REGS *outregs, struct SREGS *sregs);
void PMint86(char intnum, union REGS *inregs, union REGS *outregs);
void PMint86x(char intnum, union REGS *inregs, union REGS *outregs, struct SREGS *sregs);

int  InitVESA();
void CloseVESA();
int  VESASetGraphics(int xres, int yres, int cdepth, int lfbflag);
int  ExactModeSearch(int xres, int yres, int cdepth);
int  SetText();
void PutPixel(int x, int y, int c);
void PutPixelRGB(int x, int y, int r, int g, int b);
void Blit(char *vpage);
void BlitClear(char *vpage, int c);
void CopyWindow(char *vpage, int x1, int y1, int x2, int y2);
void SetRGB(int c, int r, int g, int b);
void SetGammaRGB(int c, int r, int g, int b, float Gamma);
int  ImageSize(int x1, int y1, int x2, int y2);
void GetImage(int x1, int y1, int x2, int y2, Sprite *Spr);
void PutImage(int x, int y, Sprite *Spr, int Style);
